/*
 * LineSensor I2C
 * version 1.1 (2011/03/15)
 *
 * (c) 2011 Adam Heinrich, <adam@adamh.cz>, www.adamh.cz
 * Latest version on: http://projects.adamh.cz/LineSensorI2C
 *
 * You may freely modify and share this code, as long as you keep this
 * notice intact (including all lines above).  Licensed under the
 * Creative Commons BY-NC-SA 3.0 license:
 *
 * http://creativecommons.org/licenses/by-nc-sa/3.0/
 *
 *
 *
 * MCU: ATmega8 @ 1 MHz
 *
 * Calibration process:
 *  1. Write 0x01 to I2C register 0x00 or hold the button
 *  2. Wait until green LED is not on without blinking
 *  3. Place all cells above the black line and press the button button OR write 0x02 to I2C register 0x00
 *  4. Place all cells above white surface and press the button OR write 0x02 to I2C register 0x00
 *  5. Calibration done, calibration values stored in EEPROM
 *
 *
 * EEPROM:
 * 	[0]    = I2C address (if 0xFF or 0x00, default address is used)
 *  [1]    = Calibration flag (if 0x01, calibration is started)
 *  [4..]  = 6*2 bytes for calibration min values
 *  [20..] = 6*2 bytes for calibration max values
 *
 *
 * I2C write registers:
 * 	[0x00] = Start calibration (if register 0x00 is set to 1)
 *  [0xAA] = I2C slave address (for 0x00 or 0xFF the default address is chosen)
 *
 *
 * I2C read registers:
 * 	[0] = Line as byte (only first 6 bits, 1 = line under the cell, 0 = no line under the cell)
 *  [1] = OnLine (1 = line under sensor, 0 = no line under sensor)
 *  [2] = Line position high byte (0-5000 -> 0 = line on the left side of the sensor, 5000 = line under the right side of the sensor)
 *  [3] = Line position low byte
 *  [4,6,8,10,12,14] = Cell value high byte (0-1000 -> 0 = white, 1000 = black)
 *  [5,7,9,11,13,15] = Cell value low byte
 *  [16,18,20,22,24,26] = Cell raw value high byte (value read from the A/D converter, not the calibrated one)
 *  [17,19,21,23,25,27] = Cell raw value low byte
 *
 */

#ifndef F_CPU
#define F_CPU 1000000  // 1 MHz
#endif

#include <avr/io.h>
#include <avr/interrupt.h>
#include <avr/wdt.h>
#include <avr/eeprom.h>
#include <util/delay.h>
#include <compat/twi.h>

// Default I2C slave address
#define DEFAULT_I2C_ADDRESS 0xaa

// A/D COnverter configuration
#define adcInit() { ADMUX = 0x40; ADCSRA = 0x87; }

// Disable WDT on startup
uint8_t mcusr_mirror __attribute__ ((section (".noinit")));

void get_mcusr(void) \
	__attribute__((naked)) \
    __attribute__((section(".init3")));
void get_mcusr(void) {
	mcusr_mirror = MCUSR;
	MCUSR = 0;
	wdt_disable();
}

// LED
#define LED_INIT()  DDRD  |=  (1<<PD0)
#define LED_ON()    PORTD |=  (1<<PD0)
#define LED_OFF()   PORTD &= ~(1<<PD0)

// Button
#define BUTTON_INIT()  PORTB  |=  (1<<PB2)
#define READ_BUTTON()    !(PINB & (1<<PB2))

volatile unsigned char i2cRegisterId = 0; // I2C register address
volatile unsigned char i2cRegisterData = 0; // I2C register data

volatile unsigned int qrdVals[6] = {0, 0, 0, 0, 0, 0}; // QRD cell values
volatile unsigned int qrdRawVals[6] = {0, 0, 0, 0, 0, 0}; // QRD raw cell values (from ADC)
volatile unsigned int linePosition = 0; // Current line position
volatile unsigned char lineByte = 0; // Current line byte
volatile unsigned char isOnLine = 0; // Current OnLine
volatile unsigned char blink = 0; // Blink flag (for blinking in main loop)
volatile unsigned char calibFlag = 0; // Calibrate Flag - button press

unsigned int qrdMin[6]; // Calibration data - min
unsigned int qrdMax[6]; // Calibration data - max
unsigned int _lastLinePosition = 0; // Last line storage

unsigned char cells[6] = { 3, 2, 1, 0, 7, 6 }; // QRD1114 cells mapped on board (ADCn input number in array)

/**
 * reads A/D Converter
 */
int readAdc(unsigned char channel) {
  unsigned int result;
  ADMUX = (channel | (0<<REFS1) | (1<<REFS0));
  ADCSRA |= 0x40;
  while (!(ADCSRA & 0x10));
  ADCSRA |= 0x10;
  result = ADCW;

  return result/2;
}

/**
 * Initialization - loads calibration values from EEPROM
 */
void loadLineCalibrationFromEEPROM() {
  unsigned char i, n;
  n = 0;
  for (i = 0; i < 12; i+=2) {
    qrdMax[n]   = eeprom_read_word((unsigned int*)(4+(i)));
    qrdMin[n++] = eeprom_read_word((unsigned int*)(20+(i)));
  }
}

/**
 * Saves minimal calib values
 */
void saveMinValues() {
  unsigned char i, n;
  n = 0;
  for (i = 0; i < 12; i+=2) {
    eeprom_write_word((unsigned int*)(20+(i)), readAdc(cells[n++]));
  }
}

/**
 * Saves maximum calib. values
 */
void saveMaxValues() {
  unsigned char i, n;
  n = 0;
  for (i = 0; i < 12; i+=2) {
	  eeprom_write_word((unsigned int*)(4+(i)), readAdc(cells[n++]));
  }
}

/**
 * Reads one QRD1114 cell
 *
 * Reading is based on calibration data, raw data is stored to qrdRawVals global array
 *
 * 0 for minimal value
 * 1000 for maximal value
 */
int readQRD1114(int i) {
  // Get raw value
  unsigned int value = readAdc(cells[i]);
  int diff = qrdMax[i]-qrdMin[i];
  qrdRawVals[i] = value;

  signed int x = 0;
  if (diff != 0) {
	 x = (((signed long) value)-qrdMin[i])*1000/diff;
  }

  // Min and max values
  if (x < 0) {
    return 0;
  } else if (x > 1000) {
	  return 1000;
  }

  return x;
}

/**
 * Reads line using weighted mean
 *
 * (Weighted mean functionality inspired by Pololu Corporation's code PololuQTRSensors.cpp (http://www.pololu.com/docs/0J18), written by Ben Schmidel et al., May 28, 2008)
 *
 * 0 = line is on left
 * 5000 = line is on right
 *
 *                   0*qrd[0] + 1000*qrd[1] +  2000*qrd[2] + ... + 5000*qrd[5]
 * linePosition  =  -----------------------------------------------------------
 *                            qrd[0] + qrd[1] + qrd[2] + ... + qrd[5]
 */
int readLine() {
  unsigned long line = 0; // long value (before division)
  unsigned long sum = 0;   // sum of qrd values
  unsigned int value = 0; // current value
  unsigned char _onLine = 0; // Is on the line?
  unsigned char _lineByte = 0; // Line Byte


  int i;
  for (i = 0; i < 6; i++) {
    // read calibrated sensor value
    value = readQRD1114(i);
    qrdVals[i] = value;

    // Hurray, one of the cells is above the line
    if (value > 200) {
      _onLine = 1;
      _lineByte |= (1<<i);
    }

    // noise filter (40 is experimental value :-) )
    if (value > 40) {
      line += (long)(value) * (i * 1000);
      sum += value;
    }
  }

  // Persist values
  lineByte = _lineByte;
  isOnLine = _onLine;
  if (_onLine) {
    // save average
    _lastLinePosition = line/sum;
  } else if (_lastLinePosition < 2500) {
    // last line was right from the center, but now we don't see any line -> return 0 value (right)
    return 0;
  } else {
    // last line was left from the center, but now we don't see any line -> return 5000 value (left)
    return 5000;
  }

  return _lastLinePosition;
}


/**
 * I2C action
 * handles global variables i2cRegisterData (read & write) and i2cRegisterAddress (read)
 */
void i2cAction(unsigned char write) {
	if (!write) { // 1 would be write
		blink = 1;
		switch (i2cRegisterId) {
			// Line byte
			case 0: i2cRegisterData = lineByte; break;

			// Is OnLine
			case 1: i2cRegisterData = isOnLine; break;

			// LinePosition (high and low byte)
			case 2: i2cRegisterData = (linePosition >> 8); break;
			case 3: i2cRegisterData = linePosition; break;

			// QRD values (high and low bytes)
			case 4: i2cRegisterData = (qrdVals[0] >> 8); break;
			case 5: i2cRegisterData = qrdVals[0]; break;
			case 6: i2cRegisterData = (qrdVals[1] >> 8); break;
			case 7: i2cRegisterData = qrdVals[1]; break;
			case 8: i2cRegisterData = (qrdVals[2] >> 8); break;
			case 9: i2cRegisterData = qrdVals[2]; break;
			case 10: i2cRegisterData = (qrdVals[3] >> 8); break;
			case 11: i2cRegisterData = qrdVals[3]; break;
			case 12: i2cRegisterData = (qrdVals[4] >> 8); break;
			case 13: i2cRegisterData = qrdVals[4]; break;
			case 14: i2cRegisterData = (qrdVals[5] >> 8); break;
			case 15: i2cRegisterData = qrdVals[5]; break;

			// QRD raw values (high and low bytes)
			case 16: i2cRegisterData = (qrdRawVals[0] >> 8); break;
			case 17: i2cRegisterData = qrdRawVals[0]; break;
			case 18: i2cRegisterData = (qrdRawVals[1] >> 8); break;
			case 19: i2cRegisterData = qrdRawVals[1]; break;
			case 20: i2cRegisterData = (qrdRawVals[2] >> 8); break;
			case 21: i2cRegisterData = qrdRawVals[2]; break;
			case 22: i2cRegisterData = (qrdRawVals[3] >> 8); break;
			case 23: i2cRegisterData = qrdRawVals[3]; break;
			case 24: i2cRegisterData = (qrdRawVals[4] >> 8); break;
			case 25: i2cRegisterData = qrdRawVals[4]; break;
			case 26: i2cRegisterData = (qrdRawVals[5] >> 8); break;
			case 27: i2cRegisterData = qrdRawVals[5]; break;
		}
	} else {
		switch (i2cRegisterId) {
			// Save calibration flag to EEPROM
			case 0x00:
				if (i2cRegisterData == 0x02) { // Emulate button press
					calibFlag = 1;
					break;
				} else if (i2cRegisterData == 0x01) { // Start calibration
					cli();
					LED_ON();
					eeprom_write_byte((unsigned char *)1, 1); // Save calibration flag
					wdt_enable(WDTO_250MS); // Restart and start calibration
					while (1);
				}
				break;

			// Change I2C slave address
			case 0xAA:
				cli();
				LED_ON();
				eeprom_write_byte(0, i2cRegisterData); // Save I2C address
				wdt_enable(WDTO_250MS); // Restart with new address
				while (1);
				break;
		}
	}
}

/**
 * I2C interrupt
 */
volatile unsigned char i2cState;
ISR(TWI_vect) {

	// TWISR (TWI status register with masked prescaler bits)
	unsigned char twiStatus = (TWSR & 0xF8);

	switch (twiStatus) {
	case TW_SR_SLA_ACK: // 0x60: SLA+W received, ACK returned
		i2cState = 0; // Start I2C state, register address required
		TWCR |= (1<<TWINT)|(1<<TWEA);
		break;

	case TW_SR_DATA_ACK: // 0x80: data received, ACK returned
		if (i2cState == 0) {
			i2cRegisterId = TWDR; // Save register address
			i2cState = 1;
		} else {
			i2cRegisterData = TWDR; // Save register data
			i2cState = 2;
		}
		TWCR |= (1<<TWINT)|(1<<TWEA);
		break;

	case TW_SR_STOP: // 0xA0: stop or repeated start condition received while selected
		if (i2cState == 2) {
			i2cAction(1);
			i2cState = 0;
		}
        TWCR |= (1<<TWINT);
		break;

	case TW_ST_SLA_ACK: // 0xA8: SLA+R received, ACK returned
	case TW_ST_DATA_ACK: // 0xB8: data transmitted, ACK received
		if (i2cState == 1) {
			i2cAction(0);
			TWDR = i2cRegisterData; // Save data to TWDR register
			i2cState = 0;
		}
		TWCR |= (1<<TWINT)|(1<<TWEA)|(1<<TWSTO);
		break;

	case TW_ST_DATA_NACK: // 0xC0: data transmitted, NACK received
        TWCR |= (1<<TWINT);
		break;

	case TW_ST_LAST_DATA: // 0xC8: last data byte transmitted, ACK received
	case TW_BUS_ERROR: // 0x00: illegal start or stop condition
	default:
        TWCR = (1<<TWEN)|(1<<TWIE)|(1<<TWINT)|(1<<TWEA)|(0<<TWSTA)|(1<<TWSTO)|(0<<TWWC);
		i2cState = 0; // Back to beginning state
	}
}




/**
 * Main Loop
 */
int main(void) {
	// Disable WDT
	wdt_disable();

	// Init I/O and ADC
	LED_INIT();
	BUTTON_INIT();
	adcInit();

	// Load I2C slave address from EEPROM or use the default one
	unsigned char i2cAddress =  eeprom_read_byte((unsigned char *)0);
	if (i2cAddress == 0 || i2cAddress == 0xff) {
		i2cAddress = DEFAULT_I2C_ADDRESS;
	}

	// Start I2C
	TWAR = i2cAddress & 0xFE; // Set I2C Address, Ignore I2C General Address 0x00
	TWDR = 0xff;
	TWCR = (1 << TWINT) | (1 << TWEA) | (1 << TWEN) | (1 << TWIE); // Start Slave Listening: Clear TWINT Flag, Enable ACK, Enable TWI, TWI Interrupt Enable

	// Enter calibration - Button pressed or calibration flag found in EEPROM
	if (READ_BUTTON() || eeprom_read_byte((unsigned char *)1) == 1) {
		if (!READ_BUTTON()) {
			sei(); // Started via I2C command, must respond to I2C interrupts
		}

		LED_ON();
		while (READ_BUTTON());
		_delay_ms(100);

		// Save BLACK into EEPROM
		while (!READ_BUTTON() && !calibFlag);
		while (READ_BUTTON());
		calibFlag = 0;
		LED_OFF();
		_delay_ms(50);
		saveMaxValues();
		LED_ON();

		// Save WHITE into EEPROM
		while (!READ_BUTTON() && !calibFlag);
		while (READ_BUTTON());
		calibFlag = 0;
		LED_OFF();
		_delay_ms(50);
		saveMinValues();
		LED_ON();
		eeprom_write_byte((unsigned char *)1, 0xff); // Clear calibration flag
		_delay_ms(100);
		LED_OFF();
	}

	loadLineCalibrationFromEEPROM();

	// Enable interrupt
	sei();

	// Startup blink
	LED_ON();
	_delay_ms(50);
	LED_OFF();
	_delay_ms(100);
	LED_ON();
	_delay_ms(50);
	LED_OFF();
	_delay_ms(100);

	unsigned int buttonTimes = 0;
	while (1) {
		linePosition = readLine();

		// Calibration trigger - button
		if (READ_BUTTON()) {
			LED_OFF();
			buttonTimes++;
			if (buttonTimes > 120) {
				cli();
				wdt_enable(WDTO_15MS);
				while (1);
			}
		} else {
			buttonTimes = 0;
			if (blink) {
				LED_ON();
				_delay_ms(2);
				LED_OFF();
				blink = 0;
			}
		}
	}

	return 0;
}
